// ------------------------------------------------
// File : channel.cpp
// Date: 4-apr-2002
// Author: giles
// Desc: 
//		Channel streaming classes. These do the actual 
//		streaming of media between clients. 
//
// (c) 2002 peercast.org
// 
// ------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "socket.h"
#include "channel.h"
#include "gnutella.h"
#include "servent.h"
#include "servmgr.h"
#include "sys.h"
#include "xml.h"
#include "http.h"

// -----------------------------------
char *Channel::srcTypes[]=
{
	"None",
	"Peercast",
	"SHOUTcast",
	"Icecast",
	"URL"
};
// -----------------------------------
char *Channel::statusMsgs[]=
{
	"NONE",
	"WAIT",
	"CONNECT",
	"REQUEST",
	"CLOSE",
	"RECEIVE",
	"BROADCAST",
	"ABORT",
	"SEARCH",
	"NOHOSTS",
	"IDLE"
};


// -----------------------------------
void readXMLString(String &str, XML::Node *n, const char *arg)
{
	char *p;
	p = n->findAttr(arg);
	if (p)
	{
		str.set(p,String::T_HTML);
		str.convertTo(String::T_ASCII);
	}
}


// -----------------------------------------------------------------------------
// Initialise the channel to its default settings of unallocated and reset.
// -----------------------------------------------------------------------------
void Channel::init()
{
	reset();
}
// -----------------------------------------------------------------------------
// Close this channel and stop thread
// -----------------------------------------------------------------------------
void Channel::close()
{
	thread.active = false;
	setStatus(S_CLOSING);
}
// -----------------------------------------------------------------------------
void Channel::endThread()
{
	close();
	thread.unlock();
	init();
}
// -----------------------------------------------------------------------------
void Channel::setStatus(STATUS s)
{
	status = s;
	if (isPlaying())
	{
		info.status = ChanInfo::S_PLAY;
		info.lastPlay = sys->getTime();
	}else{
		info.status = ChanInfo::S_UNKNOWN;
		info.lastPlay = 0;
	}

}
	
// -----------------------------------------------------------------------------
// Reset channel and make it available 
// -----------------------------------------------------------------------------
void Channel::reset()
{
	currSource.init();

	srcType = SRC_NONE;

	lastIdleTime = 0;
	prefetchCnt=0;		

	numRelays = 0;
	numListeners = 0;

	info.init();
	index = 0;

	mount.clear();
	bump = false;
	stayConnected = false;
	icyMetaInterval = 0;
	syncPos = 0;

	headMeta.init();
	insertMeta.init();

	chanData.init();

	setStatus(S_NONE);
	type = T_NONE;

	readDelay = 0;
	sock = NULL;

	sourceURL.clear();
	currComment.clear();
	currArtist.clear();
	currTitle.clear();
}

// -----------------------------------
bool	Channel::checkIdle()
{
	if ((numListeners > 0) || (stayConnected || (status == S_BROADCASTING)))
	{
		prefetchCnt=0;
		return false;
	}

	if (prefetchCnt) 
		prefetchCnt--;

	if (prefetchCnt)
		LOG_CHANNEL("prefetch %d",prefetchCnt);
	return prefetchCnt==0;
}

// -----------------------------------
bool	Channel::isFull()
{
	return chanMgr->maxStreamsPerChannel ? numRelays >= chanMgr->maxStreamsPerChannel : false;
}

// -----------------------------------
void	Channel::startMP3File(char *fn)
{
	type = T_BROADCAST;
	FileStream *fs = new FileStream();

	fs->openReadOnly(fn);
	input = fs;

	thread.data = this;
	thread.func = streamMP3File;
	if (!sys->startThread(&thread))
		init();
}
// -----------------------------------
int	Channel::streamMP3File(ThreadInfo *thread)
{
	thread->lock();

	Channel *ch = (Channel *)thread->data;

	LOG_CHANNEL("Channel started: %s",ch->getName());

	try 
	{
		while (thread->active)
		{
			ch->input->read(&ch->mp3Head,sizeof(MP3Header));

			ch->readMP3();

			ch->input->rewind();
			LOG_CHANNEL("%s end",ch->getName());
		}
	}catch(StreamException &e)
	{
		LOG_ERROR("Unable to read file: %s",e.msg);
	}
	ch->input->close();

	delete ch->input;

	LOG_CHANNEL("Channel stopped: %s",ch->getName());
	ch->endThread();
	return 0;
}
// -----------------------------------
void	Channel::startGet()
{
	srcType = SRC_PEERCAST;
	type = T_RELAY;
	input = NULL;


	thread.data = this;
	thread.func = streamGet;
	if (!sys->startThread(&thread))
		init();
}
// -----------------------------------
void	Channel::startURL(const char *u)
{
	sourceURL.set(u);

	srcType = SRC_URL;
	type = T_BROADCAST;
	stayConnected = true;

	thread.data = this;
	thread.func = streamURL;
	if (!sys->startThread(&thread))
		init();
}

// -----------------------------------
const char *Channel::getTypeStr()
{
	switch(info.type)
	{
		case ChanInfo::T_OGG:
			return "ogg";
		case ChanInfo::T_MP3:
			return "mp3";
		case ChanInfo::T_MOV:
			return "mov";
		default:
			return "stream";
	}
}
// -----------------------------------
int	Channel::findProc(ThreadInfo *thread)
{
	thread->lock();

	Channel *ch = (Channel *)thread->data;

	ch->setStatus(S_SEARCHING);

	int findCnt=0;
	while (thread->active)
	{
		ChanHitList *chl = chanMgr->findHitListByID(ch->info.id);
		if (chl && chl->numHits())
		{
			// update chaninfo with latest 
			ch->info = chl->info;

			ch->setStatus(S_IDLE);
			thread->unlock();

			ch->startGet();
			return 0;
		}else
		{
			if (findCnt == 0)
				servMgr->findChannel(ch->info);

			if (findCnt++ > 300)		// give up eventually
				break;
		}

		sys->sleep(1000);
	}

	ch->endThread();
	return 0;
}
// -----------------------------------
String Channel::streamURL(const char *url)
{
	String nextURL;

	ClientSocket *sock = NULL;

	sock = sys->createSocket();
	if (!sock)
		throw StreamException("Ch.%d cannot create socket",index);


	String urlTmp;
	urlTmp.set(url);

	char *host = urlTmp.cstr();

	if (strncmp(host,"http://",7)==0)
		host+=7;

	char *dir = strstr(host,"/");
	if (dir)
		*dir++=0;

	int port = 80;
	char *pp = strstr(host,":");
	if (pp)
	{
		port = atoi(pp+1);
		*pp=0;
	}

	LOG_CHANNEL("Fetch Host=%s",host);
	LOG_CHANNEL("Fetch Port=%d",port);
	if (dir)
		LOG_CHANNEL("Fetch Dir=%s",dir);

	//{
		try 
		{
			setStatus(S_CONNECTING);

			
			sock->open(host,port);
			sock->connect();

			HTTP http(*sock);
			http.writeLine("GET /%s HTTP/1.1",dir?dir:"");

			http.writeLine("%s %s",HTTP_HS_HOST,host);
			http.writeLine("%s %s",HTTP_HS_AGENT,PCX_AGENT);   // "WinampMPEG/2.8";
			http.writeLine("%s %s",HTTP_HS_CONNECTION,"close");
			http.writeLine("%s %s",HTTP_HS_ACCEPT,"*/*");

			http.writeLine("icy-metadata:1");

			http.writeLine("");

			int res = http.readResponse();

			if ((res!=200) && (res!=302))
			{
				LOG_ERROR("HTTP response: %s",http.cmdLine);
				throw StreamException("Bad HTTP connect");
			}

			String name = info.name;

			PlayList *pls=NULL;
			while (http.nextHeader())
			{
				LOG_CHANNEL("Fetch HTTP: %s",http.cmdLine);

				Servent::readICYHeader(http,info,NULL);
				if (http.isHeader("icy-metaint"))
					icyMetaInterval = http.getArgInt();
				else if (http.isHeader("Location:"))
					nextURL.set(http.getArgStr());

				char *arg = http.getArgStr();
				if (arg)
				{
					if (http.isHeader("content-type"))
					{
						if (stristr(arg,MIME_XSCPLS))
							pls = new PlayList(PlayList::T_SCPLS, 1000);
						else if (stristr(arg,MIME_PLS))
							pls = new PlayList(PlayList::T_PLS, 1000);
						else if (stristr(arg,MIME_TEXT))
							pls = new PlayList(PlayList::T_PLS, 1000);
					}
				}

			}

			if ((!nextURL.isEmpty()) && (res==302))
			{
				LOG_CHANNEL("Ch.%d redirect: %s",index,nextURL.cstr());
				sock->close();
				delete sock;
				sock = NULL;
				return nextURL;

			}else if (pls)
			{

				LOG_CHANNEL("Ch.%d is PLS",index);

				try
				{
					char tmp[256];
					while (sock->readLine(tmp,sizeof(tmp)))
					{
						//LOG_CHANNEL("PLS: %s",tmp);
						switch (pls->type)
						{
							case PlayList::T_PLS:
								if (tmp[0] != '#')
									pls->addURL(tmp,"");
								break;
							case PlayList::T_SCPLS:
								if (strnicmp(tmp,"file",4)==0)
								{
									char *p = strstr(tmp,"=");
									if (p)
										pls->addURL(p+1,"");
								}
								break;
						}
					}
				}catch(StreamException &)
				{}
				sock->close();
				delete sock;
				sock = NULL;

				int urlNum=0;
				String url;

				LOG_CHANNEL("Playlist: %d URLs",pls->numURLs);
				while ((thread.active) && (pls->numURLs))
				{
					if (url.isEmpty())
					{
						url = pls->urls[urlNum%pls->numURLs];
						urlNum++;
					}
					try
					{
						url = streamURL(url.cstr());
					}catch(StreamException &)
					{}
				}

				delete pls;
				
			}else
			{

				info.flags |= ChanInfo::F_RAW;

				info.id = chanMgr->broadcastID;
				info.id.encode(&servMgr->serverHost,info.name.cstr(),NULL,info.bitrate);

				input = sock;

				setStatus(S_BROADCASTING);

				//readDelay = 10;

				readStream();

				sock->close();
			}

		}catch(StreamException &e)
		{
			LOG_ERROR("Ch.%d error: %s",index,e.msg);
			//sys->sleep(3000);
		}
	//}
			
	setStatus(S_CLOSING);
	if (sock)
	{
		sock->close();
		delete sock;
	}

	return nextURL;
}

// -----------------------------------
int	Channel::streamURL(ThreadInfo *thread)
{
	thread->lock();

	Channel *ch = (Channel *)thread->data;
	ClientSocket *sock = NULL;

	LOG_CHANNEL("Ch.%d started: %s",ch->index,ch->sourceURL.cstr());


	String url;
	while (thread->active)
	{
		if (url.isEmpty())
			url = ch->sourceURL;

		url = ch->streamURL(url.cstr());
	}

	ch->endThread();

	return 0;
}	

// -----------------------------------
int	Channel::streamGet(ThreadInfo *thread)
{
	thread->lock();
	GnuPacket pack;
	Channel *ch = (Channel *)thread->data;

	chanMgr->lockHitList(ch->info.id,true);
	LOG_CHANNEL("Ch.%d started: %s",ch->index,ch->getName());

	int hptr=0;


	while (thread->active)
	{

		ch->lastIdleTime = sys->getTime();
		ch->setStatus(S_IDLE);
		while ((ch->checkIdle()) && (thread->active))
			sys->sleepIdle();

		if (!thread->active)
			break;

		//ch->info.title.set("Please wait.",String::T_ASCII);

		bool doneSearch=false;
		do 
		{
			Host sh = servMgr->serverHost;

			ChanHitList *chl = chanMgr->findHitListByID(ch->info.id);
			if (chl)
				ch->currSource = chl->getHit(&hptr,servMgr->getFirewall() == ServMgr::FW_OFF);

			if (sh.isSame(ch->currSource.host))
				ch->currSource.host.ip = 0;


			if (!ch->currSource.host.ip)
			{
				ch->setStatus(S_SEARCHING);
				if (!doneSearch)
				{
					LOG_CHANNEL("Ch.%d searching..",ch->index);
					pack.initFind(ch->info.name,NULL);
					servMgr->addReplyID(pack.id);
					servMgr->broadcast(pack,NULL);
					doneSearch = true;
				}
				sys->sleepIdle();
			}

		}while((ch->currSource.host.ip==0) && (thread->active));


		// totally give up
		if (!ch->currSource.host.ip)
			break;

		{

			try 
			{

				char hostName[64];
				ch->currSource.host.IPtoStr(hostName);


				if (!ch->currSource.firewalled || (servMgr->serverHost.localIP() && ch->currSource.host.localIP()))
				{

					ClientSocket *s = sys->createSocket();
					if (!s)
						throw StreamException("Ch.%d cannot create socket",ch->index);

					ch->sock = s;
					ch->setStatus(S_CONNECTING);
					ch->sock->open(hostName,ch->currSource.host.port);
					ch->sock->timeout = 10000;
					ch->sock->connect();

					LOG_CHANNEL("Ch.%d connect to %s",ch->index,hostName);

				}else{

					Host sh = servMgr->serverHost;


					if (!sh.isValid() || sh.loopbackIP())
						throw StreamException("No Server, unable to ask for push.");

					ch->setStatus(S_REQUESTING);

					sys->sleep(500);	// wait a bit for the previous find to go

					// wait 10 seconds for a reponse.
					int timeout = 1000;

					LOG_CHANNEL("Ch.%d Push request",ch->index);

					
					ch->pushSock = NULL;
					ch->pushIndex = ch->currSource.index;

					pack.initPush(ch->currSource,sh);
					servMgr->broadcast(pack,NULL);

					while ((!ch->pushSock) && (thread->active))
					{
						if (ch->checkBump())
							throw StreamException("Bumped");

						if (timeout-- <= 0)
							break;
						sys->sleep(1000);
					}

					if (!ch->pushSock)
						throw StreamException("Push timeout");

					ch->setStatus(S_CONNECTING);
					ch->sock = ch->pushSock;
				
				}

				char idStr[64];
				ch->info.id.toStr(idStr);
				
				if (ch->info.flags & ChanInfo::F_RAW)
				{
					// older raw data streams

					ch->sock->writeLine("GET /%s/%s HTTP/1.0",ch->getTypeStr(),idStr);

					// request metadata for mp3
					if (ch->info.type == ChanInfo::T_MP3)
						ch->sock->writeLine("icy-metadata:1");

					ch->sock->writeLine("");

					ch->icyMetaInterval = 0;
					HTTP http(*ch->sock);

					http.checkResponse(200);

					while (http.nextHeader())
					{
						if (http.isHeader("icy-metaint"))
							ch->icyMetaInterval = http.getArgInt();

						LOG_CHANNEL("Ch.%d Raw GET: %s",ch->index,http.cmdLine);
					}
				}else{

					ch->sock->writeLine("GET /channel/%s HTTP/1.0",idStr);
					ch->sock->writeLine("");
					HTTP http(*ch->sock);

					http.checkResponse(200);

					while (http.nextHeader())
						LOG_CHANNEL("Ch.%d GET: %s",ch->index,http.cmdLine);
				}

				ch->setStatus(S_RECEIVING);

				LOG_CHANNEL("Ch.%d Ready",ch->index);

				ch->input = ch->sock;

				ch->readStream();

				ch->setStatus(S_CLOSING);

			}catch(StreamException &e)
			{
				ChanHitList *chl = chanMgr->findHitListByID(ch->info.id);
				if (chl)
					chl->deadHit(ch->currSource);
				ch->setStatus(S_ABORT);
				LOG_ERROR("Ch.%d aborted: %s",ch->index,e.msg);
			}

			if (ch->sock)
			{
				ch->sock->close();
				delete ch->sock;
				ch->sock = NULL;
			}

			if (chanMgr->lastPlayID.isSame(ch->info.id))
				sys->setNotify(Sys::NT_TRACKINFO,"","");

		}

		sys->sleepIdle();
	}

	LOG_CHANNEL("Ch.%d closed",ch->index);
	chanMgr->lockHitList(ch->info.id,false);
	ch->endThread();
	return 0;
}
// -----------------------------------
void	Channel::startICY(ClientSocket *cs, SRC_TYPE st)
{
	srcType = st;
	type = T_BROADCAST;
	cs->timeout = 30000;
	sock = cs;
	input = cs;

	thread.data = this;
	thread.func = streamICY;
	if (!sys->startThread(&thread))
		init();
}
// -----------------------------------
void	Channel::startFind()
{
	type = T_RELAY;
	thread.data = this;
	thread.func = findProc;
	if (!sys->startThread(&thread))
		init();
}


// -----------------------------------
int	Channel::streamICY(ThreadInfo *thread)
{
	thread->lock();

	Channel *ch = (Channel *)thread->data;
	chanMgr->lockHitList(ch->info.id,true);


	LOG_CHANNEL("Channel started: %s",ch->getName());


	try 
	{
		while ((thread->active) && (!ch->input->eof()))
		{

			ch->setStatus(S_BROADCASTING);
			ch->readStream();
		}
		LOG_CHANNEL("Ch.%d ended",ch->index);
	}catch(StreamException &e)
	{
		ch->setStatus(S_ABORT);
		LOG_ERROR("Ch.%d aborted: %s",ch->index,e.msg);
		sys->sleep(1000);
		ch->input->close();
	}


	ch->setStatus(S_CLOSING);
	LOG_CHANNEL("Ch.%d stopped",ch->index);

	if (ch->input)
	{
		ch->input->close();
		delete ch->input;
	}

	chanMgr->lockHitList(ch->info.id,false);
	ch->endThread();
	return 0;
}
// -----------------------------------
static char *nextMetaPart(char *str,char delim)
{
	while (*str)
	{
		if (*str == delim)
		{
			*str++ = 0;
			return str;
		}
		str++;
	}
	return NULL;
}
// -----------------------------------
static void copyStr(char *to,char *from,int max)
{
	char c;
	while ((c=*from++) && (--max))
		if (c != '\'')
			*to++ = c;

	*to = 0;
}

// -----------------------------------
void Channel::processMetadata(char *str)
{
	char *cmd=str;
	while (cmd)
	{
		char *arg = nextMetaPart(cmd,'=');
		if (!arg)
			break;

		char *next = nextMetaPart(arg,';');

		if (strcmp(cmd,"StreamTitle")==0)
			info.track.title.setUnquote(arg,String::T_ASCII);
		else if (strcmp(cmd,"StreamUrl")==0)
			info.track.contact.setUnquote(arg,String::T_ASCII);


		cmd = next;
	}

	updateMeta();
}

// -----------------------------------
XML::Node *ChanHitList::createHitsXML()
{
	return new XML::Node("hits listeners=\"%d\" hosts=\"%d\" busy=\"%d\" stable=\"%d\" firewalled=\"%d\" closest=\"%d\" furthest=\"%d\" newest=\"%d\"",
		numListeners(),
		numHits(),
		numBusy(),
		numStable(),
		numFirewalled(),
		closestHit(),
		furthestHit(),
		sys->getTime()-newestHit()
		);		
}

// -----------------------------------
XML::Node *Channel::createRelayXML(bool showStat)
{
	const char *ststr;
	ststr = getStatusStr();
	if (!showStat)
		if ((status == S_RECEIVING) || (status == S_BROADCASTING))
			ststr = "OK";

	ChanHitList *chl = chanMgr->findHitListByID(info.id);

	return new XML::Node("relay listeners=\"%d\" relays=\"%d\" hosts=\"%d\" status=\"%s\"",
		numListeners,
		numRelays,
		(chl!=NULL)?chl->numHits():0,
		ststr
		);	
}

// -----------------------------------
void ChanMeta::fromXML(XML &xml)
{
	MemoryStream tout(data,MAX_DATALEN);
	xml.write(tout);

	len = tout.pos;
}
// -----------------------------------
void ChanMeta::fromMem(void *p, int l)
{
	len = l;
	memcpy(data,p,len);
}
// -----------------------------------
void ChanMeta::addMem(void *p, int l)
{
	if ((len+l) <= MAX_DATALEN)
	{
		memcpy(data+len,p,l);
		len += l;
	}
}
// -----------------------------------
void Channel::updateMeta()
{
	XML xml;

	XML::Node *n = info.createChannelXML();
	n->add(info.createTrackXML());
	n->add(info.createServentXML());

	xml.setRoot(n);

	insertMeta.fromXML(xml);

	ChanPacket pack;

	pack.init('META',insertMeta.data,insertMeta.len);
	chanData.writePacket(pack);

}
// -----------------------------------
void Channel::readStream()
{
	chanMgr->broadcastRelays(chanMgr->broadcastTTL,NULL);
	if (info.flags & ChanInfo::F_RAW)
	{
		switch(info.type)
		{
			case ChanInfo::T_MP3:
				LOG_CHANNEL("Ch.%d is MP3",index);
				readMP3();
				break;
			case ChanInfo::T_OGG:
				LOG_CHANNEL("Ch.%d is OGG",index);
				readOGG();
				break;
			case ChanInfo::T_MOV:
				LOG_CHANNEL("Ch.%d is MOV",index);
				readMOV();
				break;
			case ChanInfo::T_MPG:
				LOG_CHANNEL("Ch.%d is MPG",index);
				readMPG();
				break;
			default:
				LOG_CHANNEL("Ch.%d is Raw",index);
				readRaw();
		}
	}else{
		LOG_CHANNEL("Ch.%d is PeerCast",index);
		readPeercast();
	}

}
// -----------------------------------
void Channel::readPeercast()
{
	ChanPacket pack;

	if (input->readTag() != 'PCST')
		throw StreamException("Not PeerCast stream");

	syncPos = 0;
	info.numSkips = 0;

	while ((!input->eof() && (thread.active)))
	{

		pack.read(*input);

		chanData.writePacket(pack);
		MemoryStream mem(pack.data,pack.len);


		switch(pack.type)
		{
			case 'HEAD':
				LOG_CHANNEL("Ch.%d HEAD: %d",index,pack.len);
				if (pack.len > ChanMeta::MAX_DATALEN)
					throw StreamException("Bad HEAD");
				headMeta.fromMem(pack.data,pack.len);
				break;
			case 'META':
				LOG_CHANNEL("Ch.%d META: %d",index,pack.len);
				insertMeta.fromMem(pack.data,pack.len);
				{
					if (pack.len)
					{
						XML xml;
						xml.read(mem);
						XML::Node *n = xml.findNode("channel");					
						if (n)
						{
							info.updateFromXML(n);
							LOG_CHANNEL("Ch.%d track update: %s - %s",index,info.track.artist.cstr(),info.track.title.cstr());

							if (chanMgr->lastPlayID.isSameID(info.id))
							{
								bool commentChanged = (!currComment.isSame(info.comment) && !info.comment.isEmpty());
								bool trackChanged = (!currArtist.isSame(info.track.artist) || !currTitle.isSame(info.track.title));

								if (commentChanged || trackChanged)
								{
									sys->setNotify(Sys::NT_TRACKINFO,"","");
									sys->setNotify(Sys::NT_BROADCASTERS,"","");
								}

								if (commentChanged)
								{
									currComment = info.comment;
									String title;
									title.append("Message from ");
									title.append(info.name.cstr());
									sys->setNotify(Sys::NT_BROADCASTERS,title,info.comment.cstr());
								}

								if (trackChanged)
								{
									currArtist = info.track.artist;
									currTitle = info.track.title;
									if (!info.track.artist.isEmpty() || !info.track.title.isEmpty())
									{
										String title,msg;
										msg.append(info.name.cstr());
										msg.append(": ");
										msg.append(info.track.artist.cstr());
										msg.append(" - ");
										msg.append(info.track.title.cstr());
										sys->setNotify(Sys::NT_TRACKINFO,title.cstr(),msg.cstr());
									}
								}
							}

							ChanHitList *chl = chanMgr->findHitListByID(info.id);
							if (chl)
								chl->info.updateFromXML(n);
;
						}
					}
				}
				break;
			case 'DATA':
				//LOG_CHANNEL("DATA: %d",pack.len);
				break;
			case 'SYNC':
				{
					unsigned int s = mem.readLong();
					if ((s-syncPos) != 1)
					{
						LOG_CHANNEL("Ch.%d SKIP: %d to %d",index,syncPos,s);
						if (syncPos)
							info.numSkips++;
					}

					syncPos = s;
				}
				break;
			default:
				LOG_CHANNEL("Bad channel packet: %x",pack.type);

		}

		if (checkBump())
			break;
		if (checkIdle())
			break;

		if (readDelay)
			sys->sleep(readDelay);

	}


}
// -----------------------------------
void Channel::readRaw()
{
	ChanPacket pack;

	while ((!input->eof() && (thread.active)))
	{
		syncPos++;
		pack.init('SYNC',&syncPos,sizeof(syncPos));
		chanData.writePacket(pack);

		int rlen = 4000;

		rlen = input->read(pack.data,rlen);

		//LOG_CHANNEL("raw read %d - %d",syncPos,rlen);

		pack.len = rlen;
		pack.type = 'DATA';
		chanData.writePacket(pack);

		if (checkBump())
			break;
		if (checkIdle())
			break;

		if (readDelay)
			sys->sleep(readDelay);
	}
}

// -----------------------------------
void Channel::readMPG()
{
	ChanPacket pack;

//	for(int i=0; i<10; i++)
//	{
//		unsigned int v = in.readLong();
//		LOG_CHANNEL("raw %d: %08x",i,SWAP4(v));
//	}

	while ((!numListeners) && (thread.active))
		sys->sleep(1000);
	sys->sleep(2000);



//	in.read(headMeta.data,1024);
//	headMeta.len = 1024;

	while ((!input->eof() && (thread.active)))
	{
		syncPos++;
		pack.init('SYNC',&syncPos,sizeof(syncPos));
		chanData.writePacket(pack);

		int rlen = 4000;

		rlen = input->read(pack.data,rlen);

		//LOG_CHANNEL("raw read %d - %d",syncPos,rlen);

		pack.len = rlen;
		pack.type = 'DATA';
		chanData.writePacket(pack);

		if (checkBump())
			break;
		if (checkIdle())
			break;

		if (readDelay)
			sys->sleep(readDelay);
	}
}

// -----------------------------------
void Channel::readMP3()
{
	ChanPacket pack;

	while ((!input->eof() && (thread.active)))
	{

		if (icyMetaInterval)
		{

			int rlen = icyMetaInterval;

			while (rlen)
			{
				int rl = rlen;
				if (rl > sizeof(pack.data))
					rl = sizeof(pack.data);

				syncPos++;
				pack.init('SYNC',&syncPos,sizeof(syncPos));
				chanData.writePacket(pack);

				pack.len = rl;
				input->read(pack.data,pack.len);
				pack.type = 'DATA';
				chanData.writePacket(pack);

				rlen-=rl;
			}

			unsigned char len;
			input->read(&len,1);
			if (len)
			{
				if (len*16 > 1024) len = 1024/16;
				char buf[1024];
				input->read(buf,len*16);
				processMetadata(buf);
			}

		}else{
			input->read(pack.data,sizeof(pack.data));

			syncPos++;
			pack.init('SYNC',&syncPos,sizeof(syncPos));
			chanData.writePacket(pack);

			pack.len = sizeof(pack.data);
			pack.type = 'DATA';
			chanData.writePacket(pack);
		}

		if (checkBump())
			break;
		if (checkIdle())
			break;

		if (readDelay)
			sys->sleep(readDelay);
	}
}
// -----------------------------------
void Channel::readMOV()
{
	ChanPacket pack;

	bool done=false;
	headMeta.len = 0;
	while ((!input->eof() && (thread.active)) && (!done))
	{


		unsigned int olen = input->readLong();
		unsigned int otag = input->readLong();

        // don't ask.
		unsigned int len = SWAP4(olen);
		unsigned int tag = SWAP4(otag);

		len -= 8;


		if ((tag == 'mdat') && (len))
		{
			LOG_CHANNEL("Ch.%d mdat: %d",index,len);
			{
				LOG_CHANNEL("mov wait: %d",headMeta.len);

				headMeta.addMem(&olen,4);
				headMeta.addMem(&otag,4);

				while ((!numListeners) && (thread.active))
					sys->sleep(1000);
				sys->sleep(1000);
				LOG_CHANNEL("mov go");

				while ((len) && (thread.active))
				{
					syncPos++;
					pack.init('SYNC',&syncPos,sizeof(syncPos));
					chanData.writePacket(pack);

					int rlen = 4000;
					rlen = input->read(pack.data,rlen);
					pack.type = 'DATA';
					pack.len = rlen;

					chanData.writePacket(pack);

					len -= rlen;

					if (checkBump())
						done=true;
					if (checkIdle())
						done=true;

					if (readDelay)
						sys->sleep(readDelay);

				}
				LOG_CHANNEL("Mov data end");
			}
		}else{
			LOG_CHANNEL("Ch.%d %c%c%c%c: %d",index,tag>>24&0xff,tag>>16&0xff,tag>>8&0xff,tag&0xff,len);

			if (headMeta.len+len > ChanMeta::MAX_DATALEN)
				throw StreamException("MOV section too big");

			headMeta.addMem(&olen,4);
			headMeta.addMem(&otag,4);
			if (len)
			{
				input->read(headMeta.data+headMeta.len,len);
				headMeta.len += len;
			}
		}
	}


}
// -----------------------------------
bool OggPacket::isBOS()
{
	return (data[5] & 0x02) != 0;
}
// -----------------------------------
bool OggPacket::isVorbisPacket()
{
	char *p = (char *)getContent();

	return memcmp(p+1,"vorbis",6)==0;
}
// -----------------------------------
int OggPacket::getVorbisType()
{
	char *p = (char *)getContent();
	return p[0];
}

// -----------------------------------
void OggPacket::read(Stream &in)
{

	// skip until we get OggS identifier
	bool gotOgg=false;
	while (!gotOgg)
		if (in.readChar() == 'O')
			if (in.readChar() == 'g')
				if (in.readChar() == 'g')
					if (in.readChar() == 'S')
						gotOgg = true;


	len = 0;

	data[0] = 'O';	data[1] = 'g';	data[2] = 'g';	data[3] = 'S';

	len += 4;

	in.read(&data[len],23);
	len += 23;

	//LOG("OGG: head = %02x",data[5]);
	//LOG4("OGG: page num = %d",*(unsigned int *)&data[18]);
	//LOG("OGG: serial = %x",*(unsigned int *)&data[14]);
	numSegs = data[26];
	segLen=0;

	in.read(&data[len],numSegs);

	for(int i=0; i<numSegs; i++)
		segLen += data[len++];

	//LOG4("OGG: segs = %d/%d bytes",numSegs,segLen);
	
	if (segLen > MAX_DATALEN-128)
		throw StreamException("OGG packet too big");

	in.read(&data[len],segLen);
	len += segLen;
	//LOG4("OGG: packet len=%d",len);

}
// -----------------------------------
void ChanPacket::init(unsigned int t, const void *p, unsigned int l)
{
	type = t;
	if (l > MAX_DATALEN)
		l = MAX_DATALEN;
	len = l;
	memcpy(data,p,len);
}
// -----------------------------------
void ChanPacket::write(Stream &out)
{
	out.writeTag(type);
	out.writeShort(len);
	out.writeShort(0);
	out.write(data,len);
}
// -----------------------------------
void ChanPacket::read(Stream &in)
{
	type = in.readTag();
	len = in.readShort();
	in.readShort();
	if (len > MAX_DATALEN)
		throw StreamException("Bad ChanPacket");
	in.read(data,len);
}

#if 0
// -----------------------------------
unsigned int ChanBuffer::read(unsigned int gpos,void *p,unsigned int len)
{
	if (gpos == 0)
		gpos = pos;

	while (len)
	{
		// skip forward if too far behind
		//if ((pos > MAX_DATALEN) && (gpos < (pos-MAX_DATALEN)))
		//	gpos = pos-MAX_DATALEN;

		// sleep if no data waiting
		unsigned int tim = sys->getTime();
		while (gpos >= pos)
		{
			sys->sleepIdle();
			if ((sys->getTime() - tim) > 30)
				throw TimeoutException();
		}



		int spos = gpos % MAX_DATALEN; // start
		int epos = pos % MAX_DATALEN;	// end

		unsigned int rlen;

		if (pos >= MAX_DATALEN)
		{
			if (spos < epos)	
				rlen = epos-spos;
			else
				rlen = MAX_DATALEN-spos;
		}else{
			rlen = epos-spos;
		}

		// read as many bytes as we can
		if (rlen > len)
			rlen = len;

		if (rlen)
		{

			//LOG("BUF READ gpos=%d, pos=%d, spos=%d, epos=%d, avail=%d",gpos,pos,spos,epos,avail);

			if (rlen > MAX_DATALEN)	
			{
				// something went horribly wrong
				throw StreamException("Chan buffer read fail");
			}

			memcpy(p,&data[spos],rlen);

			p = (char *)p+rlen;
			len -= rlen;
			gpos += rlen;
		}


	}
	return gpos;
}
// -----------------------------------
void	ChanBuffer::write(const void *p,int len)
{
	while (len)
	{
		unsigned int bpos = pos % MAX_DATALEN;
		int wlen = len;
		if ((bpos+wlen) > MAX_DATALEN)
			wlen = MAX_DATALEN-bpos;
		memcpy(&data[bpos],p,wlen);

		p = (char *)p+wlen;
		len -= wlen;
		pos += wlen;
	}
}
// -----------------------------------
void	ChanBuffer::writePacket(ChanPacket &pack)
{
	lock.on();

	unsigned int t = sys->gtohl( SWAP4(pack.type) );
	
	//LOG("write packet %x,%d",t,pack.len);

	lastPacket = pos;
	
	if ((lastPacket-firstPacket) > MAX_DATALEN)
		firstPacket = lastPacket;

	write(&t,4);
	write(&pack.len,sizeof(int));
	write(pack.data,pack.len);

	lock.off();
}
#endif

// -----------------------------------
void	ChanPacketBuffer::writePacket(ChanPacket &pack)
{
	lock.on();

	packets[currPacket%MAX_PACKETS] = pack;
	lastPacket = currPacket;
	currPacket++;

	if (currPacket <= MAX_PACKETS)
		firstPacket = 0;
	else
		firstPacket = currPacket-MAX_PACKETS;

	lock.off();
}
// -----------------------------------
unsigned int	ChanPacketBuffer::readPacket(unsigned int pos,ChanPacket &pack)
{
	unsigned int tim = sys->getTime();

	if (pos > currPacket)
		pos = currPacket;

	while (pos == currPacket)
	{
		sys->sleepIdle();
		if ((sys->getTime() - tim) > 30)
			throw TimeoutException();
	}

	lock.on();

	if (pos == 0)			// start of stream
		pos = firstPacket;	// or lastPacket
	else if (pos < firstPacket)	// too far behind
		pos = firstPacket;		// so skip forward


	pack = 	packets[pos%MAX_PACKETS];
	pos++;

	lock.off();
	return pos;
}

#if 0
// -----------------------------------
unsigned int	ChanBuffer::readPacket(unsigned int gpos, ChanPacket &pack)
{
	while (gpos >= pos)
		sys->sleepIdle();

	lock.on();

	try
	{

		gpos = read(gpos,&pack.type,sizeof(pack.type));
		pack.type = sys->gtohl( SWAP4(pack.type) );

		gpos = read(gpos,&pack.len,sizeof(pack.len));

		if (pack.len > ChanPacket::MAX_DATALEN)
			pack.len = ChanPacket::MAX_DATALEN;

		gpos = read(gpos,pack.data,pack.len);
	}catch (StreamException &)
	{
	}
	lock.off();
	return gpos;
}

#endif

// -----------------------------------
void Channel::processOggIdent(OggPacket &ogg)
{
	MemoryStream in(ogg.getContent(),ogg.getContentLen());

	in.skip(7);

	int ver = in.readLong();
	int chans = in.readChar();
	int rate = in.readLong();
	int brMax = in.readLong();
	int brNom = in.readLong();
	int brLow = in.readLong();

	LOG_CHANNEL("OGG Ident: ver=%d, chans=%d, rate=%d, brMax=%d, brNom=%d, brLow=%d",
		ver,chans,rate,brMax,brNom,brLow);

	if (brNom > 0)
		info.bitrate = brNom/1000;
	else
		info.bitrate = 0;

}
// -----------------------------------
void Channel::processOggComment(OggPacket &ogg)
{
	MemoryStream in(ogg.getContent(),ogg.getContentLen());

	in.skip(7);		// skip type + 'vorbis'

	int vLen = in.readLong();	// vendor len

	in.skip(vLen);

	char argBuf[256];

	info.track.clear();

	int cLen = in.readLong();	// comment len
	for(int i=0; i<cLen; i++)
	{
		int l = in.readLong();
		if (l > sizeof(argBuf))
			throw StreamException("Comment string too long");
		in.read(argBuf,l);
		argBuf[l] = 0;
		LOG_CHANNEL("OGG Comment: %s",argBuf);

		char *arg;
		if ((arg=stristr(argBuf,"ARTIST=")))
			info.track.artist.set(arg+7,String::T_ASCII);
		else if ((arg=stristr(argBuf,"TITLE=")))
			info.track.title.set(arg+6,String::T_ASCII);
		else if ((arg=stristr(argBuf,"GENRE=")))
			info.track.genre.set(arg+6,String::T_ASCII);
		else if ((arg=stristr(argBuf,"CONTACT=")))
			info.track.contact.set(arg+8,String::T_ASCII);
		else if ((arg=stristr(argBuf,"ALBUM=")))
			info.track.album.set(arg+6,String::T_ASCII);
	}

	updateMeta();

}

// -----------------------------------
void Channel::getStreamPath(char *str)
{
	char idStr[64];

	getIDStr(idStr);

	sprintf(str,"/stream/%s.%s",idStr,getTypeStr());
}
// -----------------------------------
void Channel::readOGG()
{

	OggPacket ogg;
	ChanPacket pack;

	while ((!input->eof() && (thread.active)))
	{

		ogg.read(*input);


		if (ogg.isBOS())
		{
			LOG_CHANNEL("Got OGG BOS");
			headMeta.len = 0;
			headMeta.cnt = 0;
		}

		if (ogg.isVorbisPacket())
		{
			if (ogg.getVorbisType()==3)
			{
				LOG_CHANNEL("Vorbis: Comment");
				processOggComment(ogg);
			}
		}

		// first 3 headers of OGG are needed
		if (headMeta.cnt < 3)		
		{
			//LOG("New OGG Header %d - %02x - %x",headMeta.cnt,ogg.data[5],*(unsigned int *)&ogg.data[14]);

			if (headMeta.len+ogg.len > ChanMeta::MAX_DATALEN)
				throw StreamException("OGG packet too big for headMeta");

			if (ogg.isVorbisPacket())
			{

				switch (ogg.getVorbisType())
				{
					case 1:	// ident
						LOG_CHANNEL("Vorbis Header: Ident");
						processOggIdent(ogg);
						break;
					case 3: // comment
						LOG_CHANNEL("Vorbis Header: Comment");
						//processOggComment(ogg);
						break;
					case 5: // setup
						LOG_CHANNEL("Vorbis Header: Setup");
						break;
				}

				memcpy(&headMeta.data[headMeta.len],ogg.data,ogg.len);
				headMeta.len += ogg.len;
				headMeta.cnt++;

			}else{
				// we`ve only had 2 vorbis headers, but force a continue anyway.
				headMeta.cnt = 3;
				LOG_CHANNEL("Vorbis: No Setup header, continue anyway...");
			}

			if (headMeta.cnt == 3)
			{
				pack.init('HEAD',headMeta.data,headMeta.len);
				chanData.writePacket(pack);
			}
		}

		syncPos++;
		pack.init('SYNC',&syncPos,sizeof(syncPos));
		chanData.writePacket(pack);

		pack.init('DATA',ogg.data,ogg.len);
		chanData.writePacket(pack);

		if (checkBump())
			break;
		if (checkIdle())
			break;

		if (readDelay)
			sys->sleep(readDelay);

	}
}
// -----------------------------------
void ChanMgr::startSearch(ChanInfo &info)
{
	searchInfo = info;
	clearHitLists();
	numFinds = 0;
	lastHit = 0;
	lastSearch = 0;
	searchActive = true;
}
// -----------------------------------
void ChanMgr::lockHitList(GnuID &id, bool on)
{
	ChanHitList *chl = findHitListByID(id);
	if (chl)
		chl->locked = on;
}

// -----------------------------------
Channel *ChanMgr::findChannel(ChanInfo &info)
{
	Channel *c=NULL;
	findChannels(info,&c,1);
	return c;
}

// -----------------------------------
Channel *ChanMgr::findChannelByName(const char *n)
{
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (stricmp(channels[i].info.name,n)==0)
				return &channels[i];

	return NULL;
}
// -----------------------------------
Channel *ChanMgr::findChannelByIndex(int index)
{
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].index == index)
				return &channels[i];

	return NULL;
}	
// -----------------------------------
Channel *ChanMgr::findChannelByMount(const char *str)
{
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (strcmp(channels[i].mount,str)==0)
				return &channels[i];

	return NULL;
}	
// -----------------------------------
Channel *ChanMgr::findChannelByID(GnuID &id)
{
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].info.id.isSameID(id))
				return &channels[i];

	return NULL;
}	
// -----------------------------------
Channel *ChanMgr::findPushChannel(int index)
{
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if ((channels[i].pushIndex == index) && (!channels[i].pushSock))
				return &channels[i];

	return NULL;
}	
// -----------------------------------
int ChanMgr::findChannels(ChanInfo &info, Channel **ch, int max)
{
	int cnt=0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].info.match(info))
			{
				ch[cnt++] = &channels[i];
				if (cnt >= max)
					break;
			}
	return cnt;
}
// -----------------------------------
Channel *ChanMgr::createRelay(ChanInfo &info, bool stayConnected)
{
	Channel *c = chanMgr->createChannel(info,NULL);
	if (c)
	{
		c->stayConnected = stayConnected;
		c->startFind();
		return c;
	}
	return NULL;
}
// -----------------------------------
int ChanMgr::findAndRelay(ChanInfo &info, Channel **ch, int max)
{
	int cnt=0;

	for(int i=0; i<30; i++)	
	{
		LOG_CHANNEL("Searching for (%d): %s",i,info.name.cstr());

		ChanHitList *chl = findHitList(info);

		if (chl)
		{
			Channel *c;

			c = findChannelByID(chl->info.id);

			if (!c)
			{
				c = chanMgr->createChannel(chl->info,NULL);
				if (c)
					c->startGet();
			}

			if (!c)
				break;
			
			ch[0] = c;
			cnt = 1;
			break;
		}else
		{
			if (i == 0)
				servMgr->findChannel(info);
		}

		sys->sleep(1000);
	}

	if (!cnt)
		LOG_CHANNEL("Not found: %d",cnt);

	return cnt;
}
// -----------------------------------
ChanMgr::ChanMgr()
{
	int i;

	for(i=0; i<MAX_CHANNELS; i++)
		channels[i].init();


	for(i=0; i<MAX_HITLISTS; i++)
	{
		hitlists[i].index = i;
		hitlists[i].init();
	}

	broadcastMsg.clear();
	broadcastMsgInterval=10;

	broadcastID.generate();

	deadHitAge = 300;

	icyMetaInterval = 8192;	
	maxStreamsPerChannel = 0;

	searchInfo.init();
	fullHitFrequency = 1;
	broadcastTTL = 7;

	lastPlayID.clear();
}
// -----------------------------------
void ChanMgr::broadcastRelays(int ttl, Servent *serv)
{
	//if ((servMgr->getFirewall() == ServMgr::FW_OFF) || servMgr->serverHost.localIP())
	{
		Channel *hits[16];
		int cnt=0;

		ChanInfo info;
		info.init();
		info.status = ChanInfo::S_PLAY;
		cnt = findChannels(info,hits,16);

		GnuPacket out;
		if (cnt)
		{
			Host sh = servMgr->serverHost;
			bool push = (servMgr->getFirewall()==ServMgr::FW_ON);
			bool busy = (servMgr->inFull() && servMgr->outFull()) || servMgr->streamFull();
			bool stable = servMgr->totalStreams>0;
			GnuID rid = out.initHit(sh,cnt,hits,NULL,0x0114,push,busy,stable,ttl);
			servMgr->addReplyID(rid);

			int numOut=0;
			if (serv)
			{
				if (serv->outputPacket(out))
					numOut=1;
			}else
				numOut = servMgr->broadcast(out,NULL);

			LOG_CHANNEL("Relay Broadcast %d channels to %d servents",cnt,numOut);		
		}
	}
}
// -----------------------------------
void ChanMgr::setBroadcastMsg(String &msg)
{
	//if (!msg.isSame(broadcastMsg))
	{
		broadcastMsg = msg;

		for(int i=0; i<MAX_CHANNELS; i++)
		{
			Channel *c = &channels[i];
			if (c->isActive())
				if (c->status == Channel::S_BROADCASTING)
				{
					c->info.comment = broadcastMsg;
					c->updateMeta();
				}
		}

	}
}
// -----------------------------------
void ChanMgr::clearHitLists()
{
	for(int i=0; i<MAX_HITLISTS; i++)
		if (!hitlists[i].locked)
			hitlists[i].init();
}

// -----------------------------------
Channel *ChanMgr::createChannel(ChanInfo &info, const char *mount)
{
	lock.on();
	for(int i=0; i<MAX_CHANNELS; i++)
	{
		Channel *c = &channels[i];
		if (!c->isActive())
		{
			c->info = info;
			c->info.lastPlay = 0;
			c->info.status = ChanInfo::S_UNKNOWN;
			if (mount)
				c->mount.set(mount);
			c->index = i+1;
			c->setStatus(Channel::S_WAIT);
			c->type = Channel::T_ALLOCATED;
			lock.off();
			return c;
		}
	}
	lock.off();
	return NULL;
}
// -----------------------------------
ChanHitList *ChanMgr::findHitList(ChanInfo &info)
{
	for(int i=0; i<MAX_HITLISTS; i++)
		if (hitlists[i].isUsed())
			if (hitlists[i].info.match(info))
				return &hitlists[i];
	return NULL;
}
// -----------------------------------
ChanHitList *ChanMgr::findHitListByID(GnuID &id)
{
	for(int i=0; i<MAX_HITLISTS; i++)
		if (hitlists[i].isUsed())
			if (hitlists[i].info.id.isSame(id))
				return &hitlists[i];
	return NULL;
}
// -----------------------------------
int ChanMgr::numHitLists()
{
	int num=0;
	for(int i=0; i<MAX_HITLISTS; i++)
		if (hitlists[i].isUsed())
			num++;
	return num;
}
// -----------------------------------
ChanHitList *ChanMgr::addHitList(ChanInfo &info)
{
	ChanHitList *hl = NULL;


	for(int i=0; i<MAX_HITLISTS; i++)
		if (!hitlists[i].isUsed())
		{
			hl = &hitlists[i];
			break;
		}

	if (hl)
		hl->info = info;

	return hl;
}
// -----------------------------------
void ChanMgr::clearDeadHits()
{
	for(int i=0; i<MAX_HITLISTS; i++)
		if (hitlists[i].isUsed())
			if (hitlists[i].clearDeadHits(deadHitAge) == 0)
				hitlists[i].init();
}

// -----------------------------------
int ChanMgr::numConnected()
{
	int tot = 0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			tot++;
	return tot;
}
// -----------------------------------
int ChanMgr::numRelayed()
{
	int tot = 0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].isPlaying())
				if (channels[i].numRelays>0)
					tot++;
	return tot;
}
// -----------------------------------
int ChanMgr::numListeners()
{
	int tot = 0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].isPlaying())
				if (channels[i].numListeners>0)
					tot++;
	return tot;
}
// -----------------------------------
int ChanMgr::numIdle()
{
	int tot = 0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].status == Channel::S_IDLE)
				tot++;
	return tot;
}

// -----------------------------------
unsigned int ChanMgr::totalInput()
{
	unsigned int tot = 0;
	for(int i=0; i<MAX_CHANNELS; i++)
		if (channels[i].isActive())
			if (channels[i].isPlaying())
				if (channels[i].input)
					tot+=channels[i].input->bytesInPerSec;

	return tot;
}
// -----------------------------------
ChanHit	*ChanMgr::addHit(ChanInfo &info, ChanHit &h)
{
	if (searchActive)
		lastHit = sys->getTime();

	// if channel is waiting for update then give it
	Channel *ch = findChannelByID(info.id);
	if (ch)
		if (!info.name.isEmpty())
			ch->info.update(info);


	// otherwise add to list of channels
	ChanHitList *hl = findHitListByID(info.id);
	if (!hl)
		hl = addHitList(info);

	if (hl)
	{
		if (!info.name.isEmpty())
			hl->info.update(info);
		return hl->addHit(h);
	}else
		return NULL;
}

// -----------------------------------
void ChanMgr::playChannels(Channel **cl, int num)
{
	if (num)
	{
		char str[128];
		sprintf(str,"http://localhost:%d",servMgr->serverHost.port);

		PlayList *pls = new PlayList(PlayList::T_SCPLS,num);
		pls->addChannels(str,cl,num);


		FileStream file;
		file.openWriteReplace("play.pls");
		pls->write(file);
		file.close();

		sys->executeFile("play.pls");

		delete pls;
	}

}
// -----------------------------------
class ChanFindInfo : public ThreadInfo
{
public:
	GnuID	id;
	bool	keep;
};
// -----------------------------------
THREAD_PROC playChannelProc(ThreadInfo *th)
{
	ChanFindInfo *cfi = (ChanFindInfo *)th;

	ChanInfo info;
	info.id = cfi->id;

	sys->setNotify(Sys::NT_PEERCAST,"PeerCast","Connecting to channel, please wait...");

	Channel *ch;
	int num = chanMgr->findChannels(info,&ch,1);

	if (!num)
		num = chanMgr->findAndRelay(info,&ch,1);

	if (num)
	{
		ch->prefetchCnt = 100;	// prefetch for 100 packets before giving up and closing channel
		chanMgr->playChannels(&ch,1);
		ch->stayConnected = cfi->keep;
	}

	delete cfi;
	return 0;
}
// -----------------------------------
void ChanMgr::playChannel(GnuID &id, bool keep)
{
	ChanFindInfo *cfi = new ChanFindInfo;
	cfi->id = id;
	cfi->keep = keep;
	cfi->func = playChannelProc;
	lastPlayID = id;
	sys->startThread(cfi);
}
// -----------------------------------
void ChanHitList::init()
{
	info.init();
	memset(hits,0,sizeof(ChanHit)*MAX_HITS);
	lastHitTime = 0;
	locked = false;
}
// -----------------------------------
bool	ChanHitList::isAvailable() 
{
	return (numHits()-numBusy())>0;
}

// -----------------------------------
ChanHit *ChanHitList::addHit(ChanHit &h)
{
	int i;

	lastHitTime = sys->getTime();
	h.time = lastHitTime;

	for(i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip == h.host.ip)
			if (hits[i].host.port == h.host.port)
			{
				hits[i] = h;
				return &hits[i];
			}


	for(i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip == 0)
		{
			hits[i] = h;
			return &hits[i];
		}

	return NULL;
}


// -----------------------------------
int	ChanHitList::clearDeadHits(unsigned int timeout)
{
	int cnt=0;
	unsigned int ctime = sys->getTime();
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
		{
			if ((ctime-hits[i].time) > timeout)
				hits[i].init();
			else
				cnt++;
		}
	return cnt;
}


// -----------------------------------
void	ChanHitList::deadHit(ChanHit &h)
{
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip == h.host.ip)
			if (hits[i].host.port == h.host.port)
				hits[i].init();
}
// -----------------------------------
int	ChanHitList::numHits()
{
	int cnt=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			cnt++;

	return cnt;
}
// -----------------------------------
int	ChanHitList::numListeners()
{
	int cnt=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			cnt += hits[i].numListeners;

	return cnt;
}
// -----------------------------------
int	ChanHitList::numBusy()
{
	int cnt=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			cnt += hits[i].busy?1:0;

	return cnt;
}
// -----------------------------------
int	ChanHitList::numStable()
{
	int cnt=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			cnt += hits[i].stable?1:0;

	return cnt;
}
// -----------------------------------
int	ChanHitList::numFirewalled()
{
	int cnt=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			cnt += hits[i].firewalled?1:0;

	return cnt;
}
// -----------------------------------
int	ChanHitList::closestHit()
{
	int hop=10000;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			if (hits[i].hops < hop)
				hop = hits[i].hops;

	return hop;
}
// -----------------------------------
int	ChanHitList::furthestHit()
{
	int hop=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			if (hits[i].hops > hop)
				hop = hits[i].hops;

	return hop;
}
// -----------------------------------
unsigned int	ChanHitList::newestHit()
{
	unsigned int time=0;
	for(int i=0; i<MAX_HITS; i++)
		if (hits[i].host.ip)
			if (hits[i].time > time)
				time = hits[i].time;

	return time;
}


// -----------------------------------
ChanHit	ChanHitList::getHit(int *ptr, bool allowPush)
{
	for(int i=0; i<MAX_HITS; i++)
	{
		int idx = (*ptr)++;
		if (idx >= MAX_HITS)
		{
			*ptr = 0;
			idx = 0;
		}
		if ((allowPush && hits[idx].firewalled) || (!hits[idx].firewalled) || (hits[idx].host.localIP()))
			if (hits[idx].host.ip)
				if (!hits[idx].busy)
					return hits[idx];
	}
	ChanHit ch;
	ch.init();
	return ch;
}

// -----------------------------------
void ChanHit::init()
{
	host.ip = 0;
	host.port = 0;
	numListeners = 0;
	firewalled = busy = stable = false;
	hops = 0;
	index = 0;
	time = upTime = 0;
	agentStr[0]=0;
	numSkips=0;
}
// -----------------------------------
const char *ChanInfo::getTypeStr()
{
	switch (type)
	{
		case T_OGG: return "OGG";
		case T_MP3: return "MP3";
		case T_MOV: return "MOV";
		case T_MPG: return "MPG";
		default: return "Unknown";
	}
}
// -----------------------------------
ChanInfo::TYPE ChanInfo::getTypeFromStr(const char *str)
{
	if (stricmp(str,"MP3")==0)
		return T_MP3;
	else if (stricmp(str,"OGG")==0)
		return T_OGG;
	else 
		return T_UNKNOWN;
}
// -----------------------------------
bool	ChanInfo::match(ChanInfo &inf)
{
	bool matchAny=true;

	if (inf.status != S_UNKNOWN)
	{
		if (status == inf.status)
			return true;
		else
			return false;
	}

	if (inf.bitrate != 0)
	{
		if (bitrate == inf.bitrate)
			return true;
		matchAny = false;
	}

	if (inf.id.isSet())
	{
		if (id.isSame(inf.id))
			return true;
		matchAny = false;
	}

	if (inf.type != T_UNKNOWN)
	{
		if (type == inf.type)
			return true;
		matchAny = false;
	}

	if (!inf.name.isEmpty())
	{
		if (name.contains(inf.name))
			return true;
		matchAny = false;
	}

	if (!inf.genre.isEmpty())
	{
		if (genre.contains(inf.genre))
			return true;
		matchAny = false;
	}

	return matchAny;
}

// -----------------------------------
void ChanInfo::update(ChanInfo &info)
{
	//if (type == T_UNKNOWN)
	{
		bitrate = info.bitrate;
		type = info.type;
		track = info.track;
		url = info.url;
		comment = info.comment;
		genre = info.genre;
	}
}
// -----------------------------------
void ChanInfo::initNameID(const char *n)
{
	init();
	id.fromStr(n);
	if (!id.isSet())
		name.set(n);
}

// -----------------------------------
void ChanInfo::init()
{
	status = S_UNKNOWN;
	name.clear();
	bitrate = 0;
	type = T_UNKNOWN;
	flags = 0;
	id.clear();
	url.clear();
	genre.clear();
	comment.clear();
	track.clear();
	lastPlay = 0;
	numSkips = 0;

}
// -----------------------------------
void ChanInfo::readTrackXML(XML::Node *n)
{
	track.clear();
	readXMLString(track.title,n,"title");
	readXMLString(track.contact,n,"contact");
	readXMLString(track.artist,n,"artist");
	readXMLString(track.album,n,"album");
	readXMLString(track.genre,n,"genre");
}
// -----------------------------------
XML::Node *ChanInfo::createChannelXML()
{
	char idStr[64];

	String nameHTML = name;
	nameHTML.convertTo(String::T_HTML);
	String urlHTML = url;
	urlHTML.convertTo(String::T_HTML);
	String genreHTML = genre;
	genreHTML.convertTo(String::T_HTML);
	String descHTML = desc;
	descHTML.convertTo(String::T_HTML);
	String commentHTML;
	commentHTML = comment;
	commentHTML.convertTo(String::T_HTML);


	id.toStr(idStr);


	return new XML::Node("channel name=\"%s\" id=\"%s\" bitrate=\"%d\" type=\"%s\" genre=\"%s\" desc=\"%s\" url=\"%s\" uptime=\"%d\" comment=\"%s\" skips=\"%d\"",
		nameHTML.cstr(),
		idStr,
		bitrate,
		getTypeStr(),
		genreHTML.cstr(),
		descHTML.cstr(),
		urlHTML.cstr(),
		lastPlay?(sys->getTime()-lastPlay):0,
		commentHTML.cstr(),
		numSkips
		);	
}
// -----------------------------------
XML::Node *ChanInfo::createRelayChannelXML()
{
	char idStr[64];

	id.toStr(idStr);


	return new XML::Node("channel id=\"%s\" uptime=\"%d\" skips=\"%d\"",
		idStr,
		lastPlay?(sys->getTime()-lastPlay):0,
		numSkips
		);	
}// -----------------------------------
XML::Node *ChanInfo::createTrackXML()
{
	String titleHTML,contactHTML,albumHTML,genreHTML,artistHTML;
	titleHTML = track.title;
	titleHTML.convertTo(String::T_HTML);
	artistHTML = track.artist;
	artistHTML.convertTo(String::T_HTML);
	albumHTML = track.album;
	albumHTML.convertTo(String::T_HTML);
	genreHTML = track.genre;
	genreHTML.convertTo(String::T_HTML);
	contactHTML = track.contact;
	contactHTML.convertTo(String::T_HTML);


	return new XML::Node("track title=\"%s\" artist=\"%s\" album=\"%s\" genre=\"%s\" contact=\"%s\"",
		titleHTML.cstr(),
		artistHTML.cstr(),
		albumHTML.cstr(),
		genreHTML.cstr(),
		contactHTML.cstr()
		);
}
// -----------------------------------
XML::Node *ChanInfo::createServentXML()
{

	return new XML::Node("servent agent=\"%s\"",
		PCX_AGENT
		);
}


// -----------------------------------
String ChanInfo::createFullNameID()
{
	String fn;

	char idStr[64];
	id.toStr(idStr);

	fn = name;
	fn.append(" :ID");
	fn.append(idStr);

	if (bitrate)
	{
		char brstr[64];
		sprintf(brstr," :BR%d",bitrate);
		fn.append(brstr);
	}

	// add channel type
	switch(type)
	{
		case ChanInfo::T_OGG:
			fn.append(" :DTOGG");
			break;
		case ChanInfo::T_MP3:
			fn.append(" :DTMP3");
			break;
	}

	return fn;
}

// -----------------------------------
void ChanInfo::init(XML::Node *n)
{
	init();

	updateFromXML(n);
}
// -----------------------------------
void ChanInfo::updateFromXML(XML::Node *n)
{
	String typeStr,idStr;

	readXMLString(name,n,"name");
	readXMLString(genre,n,"genre");
	readXMLString(url,n,"url");
	readXMLString(desc,n,"desc");


	int br = n->findAttrInt("bitrate");
	if (br)
		bitrate = br;

	readXMLString(typeStr,n,"type");
	if (!typeStr.isEmpty())
		type = getTypeFromStr(typeStr.cstr());


	readXMLString(idStr,n,"id");
	if (!idStr.isEmpty())
		id.fromStr(idStr.cstr());

	readXMLString(comment,n,"comment");

	XML::Node *tn = n->findNode("track");
	if (tn)
		readTrackXML(tn);

}

// -----------------------------------
void ChanInfo::init(const char *n, GnuID &cid, TYPE tp, int br)
{
	init();

	name.set(n);
	bitrate = br;
	type = tp;
	id = cid;
}

// -----------------------------------
void ChanInfo::init(const char *fn)
{
	init();

	if (fn)
	{
		if (strstr(fn,":DTOGG"))
			type = T_OGG;
		else if (strstr(fn,":DTMP3"))
			type = T_MP3;

		char *brs;
		if ((brs=strstr(fn,":BR"))!=NULL)
		{
			String br;
			br.setFromWord(brs+3);
			bitrate = atoi(br.cstr());
		}

		name.set(fn);
		char *idp = strstr(name.cstr(),":ID");
		if (idp)
		{
			id.fromStr(idp+3);
			idp[-1] = 0;
		}

	}
}

// -----------------------------------
void PlayList::writeSCPLS(Stream &out)
{
	out.writeLine("[playlist]");
	out.writeLine("");
	out.writeLine("NumberOfEntries=%d",numURLs);

	for(int i=0; i<numURLs; i++)
	{
		out.writeLine("File%d=%s",i+1,urls[i].cstr());
		out.writeLine("Title%d=%s",i+1,titles[i].cstr());
		out.writeLine("Length%d=-1",i+1);
	}
}
// -----------------------------------
void PlayList::writePLS(Stream &out)
{
	for(int i=0; i<numURLs; i++)
		out.writeLine("%s",urls[i].cstr());
}

// -----------------------------------
void PlayList::addChannels(const char *path, Channel **cl, int num)
{
	String url;

	for(int i=0; i<num; i++)
	{
		char idStr[64];

		cl[i]->getIDStr(idStr);

		sprintf(url.cstr(),"%s/stream/%s.%s",path,idStr,cl[i]->getTypeStr());
		addURL(url.cstr(),cl[i]->info.name);

	}	
}
