// ------------------------------------------------
// File : servent.cpp
// Date: 4-apr-2002
// Author: giles
// Desc: 
//		Servents are the actual connections between clients. They do the handshaking,
//		transfering of data and processing of GnuPackets. Each servent has one socket allocated
//		to it on connect, it uses this to transfer all of its data.
//
// (c) 2002 peercast.org
// ------------------------------------------------
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------


#include <stdlib.h>
#include "servent.h"
#include "sys.h"
#include "gnutella.h"
#include "xml.h"
#include "html.h"
#include "http.h"
#include "stats.h"
#include "servmgr.h"



// -----------------------------------
char *Servent::statusMsgs[]=
{
        "NONE",
		"CONNECTING",
        "PROTOCOL",
        "HANDSHAKE",
        "CONNECTED",
        "CLOSING",
		"LISTENING",
		"TIMEOUT",
		"REFUSED",
		"VERIFIED",
		"ERROR",
		"WAIT"
};

// -----------------------------------
char *Servent::typeMsgs[]=
{
		"NONE",
		"ALLOCATED",
        "OUT",
        "IN",
        "SERVER",
		"CHECK",
		"STREAM",
        "LOOKUP"
};

// -----------------------------------
bool	Servent::isPrivate() 
{
	Host h = getHost();
	return servMgr->isFiltered(ServFilter::F_PRIVATE,h);
}
// -----------------------------------
bool	Servent::isAllowed(int a) 
{
	Host h = getHost();

	if (!h.isValid())
		return false;

	if (servMgr->isFiltered(ServFilter::F_BAN,h))
		return false;

	if (!servMgr->isFiltered(ServFilter::F_ALLOW,h))
		return false;

	return (allow&a)!=0;
}

// -----------------------------------
void Servent::init()
{
	seenIDs.init(MAX_HASH);
	permAlloc = false;
	sock = NULL;
	reset();
}
// -----------------------------------
void Servent::reset(bool doClose)
{

	if (doClose)
		close();

	agent.clear();
	sock = NULL;
	outputBitrate = 0;
	chanIndex = -1;
	allow = ALLOW_ALL;
	currPos = 0;
	addMetadata = false;
	pack.func = 255;
	lastConnect = lastPing = lastPacket = 0;
	loginPassword[0] = 0;
	loginMount[0] = 0;
	isRaw = false;
	bytesPerSecond = 0;
	readOut = writeOut = 0;
	seenIDs.clearAll();

	status = S_NONE;
	if (!permAlloc)
		type = T_NONE;
}
// -----------------------------------
Host Servent::getHost()
{
	Host h(0,0);

	if (sock)
		h = sock->host;

	return h;
}

// -----------------------------------
bool Servent::outputPacket(GnuPacket &p)
{
	if ((writeOut-readOut) < MAX_OUTPACKETS)
	{
		outPack[writeOut%MAX_OUTPACKETS] = p;
		writeOut++;
		return true;
	}else
	{
		return false;
	}
}

// -----------------------------------
bool Servent::initServer(Host &h)
{
	try
	{
		checkFree();

		status = S_WAIT;

		createSocket();

		sock->bind(h);

		thread.data = this;

		thread.func = serverProc;

		type = T_SERVER;

		if (!sys->startThread(&thread))
			throw StreamException("Can`t start thread");

	}catch(StreamException &e)
	{
		LOG_ERROR("Bad server: %s",e.msg);
		reset();
		return false;
	}

	return true;
}
// -----------------------------------
void Servent::checkFree()
{
	if (sock)
		throw StreamException("Socket already set");
	if (thread.active)
		throw StreamException("Thread already active");
}
// -----------------------------------
void Servent::initIncoming(ClientSocket *s, unsigned int a)
{

	try{

		checkFree();

		type = T_INCOMING;
		sock = s;
		allow = a;
		thread.data = this;
		thread.func = incomingProc;

		setStatus(S_PROTOCOL);

		if (!sys->startThread(&thread))
		{
			reset();
			LOG_ERROR("Unable to start incoming thread");
		}
	}catch(StreamException &e)
	{
		LOG_ERROR("Incoming error: %s",e.msg);
		reset();
	}
}

// -----------------------------------
void Servent::initOutgoing(char *hostName, int port, TYPE ty)
{
	try 
	{
		checkFree();

	    createSocket();

		type = ty;

		sock->timeout = 10000;	// 10 seconds to connect 
		sock->open(hostName,port);

		if (!isAllowed(ALLOW_SERVENT))
			throw StreamException("Servent not allowed");

		thread.data = this;
		thread.func = outgoingProc;

		if (!sys->startThread(&thread))
			throw StreamException("Can`t start thread");

	}catch(StreamException &e)
	{
		LOG_ERROR("Unable to open connection to %s:%d - %s",hostName,port,e.msg);
		reset();
	}
}
// -----------------------------------
void Servent::initGiv(Host &h, int idx, GnuID &id)
{
	char hostName[64];
	h.IPtoStr(hostName);
	try 
	{

		checkFree();


		Channel *ch = chanMgr->findChannelByIndex(idx);
		if (!ch)
			throw StreamException("No channel");

		chanInfo = ch->info;
		chanIndex = idx;
		pushID	= id;

	    createSocket();

		sock->open(hostName,h.port);

		if (!isAllowed(ALLOW_DATA))
			throw StreamException("Servent not allowed");
		
		sock->connect();

		thread.data = this;
		thread.func = givProc;

		type = T_STREAM;

		//isRaw = (ch->info.flags & ChanInfo::F_RAW)!=0;
		isRaw = false;

		if (!sys->startThread(&thread))
			throw StreamException("Can`t start thread");

	}catch(StreamException &e)
	{
		LOG_ERROR("Giv error: 0x%x to %s: %s",idx,hostName,e.msg);
		reset();
	}
}
// -----------------------------------
void Servent::createSocket()
{
	if (sock)
		LOG_ERROR("Servent::createSocket attempt made while active");

	sock = sys->createSocket();
}
// -----------------------------------
void Servent::endThread()
{	
	reset(true);
	thread.unlock();
}
// -----------------------------------
void Servent::abort()
{
	thread.active = false;
	if (sock)
		sock->close();
}
// -----------------------------------
void Servent::close()
{	
	thread.active = false;
		
	setStatus(S_CLOSING);
	if (sock)
	{
		sock->close();
		delete sock;
		sock = NULL;
	}
}

// -----------------------------------
void Servent::setStatus(STATUS s)
{
	if (s != status)
	{
		status = s;

		if ((s == S_CONNECTED) || (s == S_LISTENING))
			lastConnect = sys->getTime();
	}

}
// -----------------------------------
void Servent::handshakeOut()
{
	sock->timeout = 10000;

	if (servMgr->allowGnutella)
	    sock->writeLine(GNU_CONNECT);
	else
	    sock->writeLine(GNU_PEERCONN);
    
	sock->writeLine("%s %s",HTTP_HS_AGENT,PCX_AGENT);
    sock->writeLine("%s %s",PCX_HS_SUBNET,servMgr->network.cstr());

	char str[64];
	servMgr->sessionID.toStr(str);
	sock->writeLine("%s %s",PCX_HS_ID,str);

	
#if _LINUX

#if _LINUX_STATIC
    sock->writeLine("%s %s",PCX_HS_OS,PCX_OS_LINUXSTA);
#else
    sock->writeLine("%s %s",PCX_HS_OS,PCX_OS_LINUXDYN);
#endif

#endif

#if WIN32
    sock->writeLine("%s %s",PCX_HS_OS,PCX_OS_WIN32);
#endif

#ifdef __APPLE__
    sock->writeLine("%s %s",PCX_HS_OS,PCX_OS_MACOSX);
#endif

	sock->writeLine("");

	HTTP http(*sock);

	http.checkResponse(200);

	bool subnetValid = false;
	bool versionValid = false;

    while (http.nextHeader())
    {
		char *arg = http.getArgStr();
		if (!arg)
			continue;

		if (http.isHeader(PCX_HS_SUBNET))
			subnetValid = stricmp(arg,servMgr->network.cstr())==0;
		else if (http.isHeader(HTTP_HS_AGENT))
		{
			agent.set(arg);

			if (strnicmp(arg,"PeerCast/",9)==0)
				versionValid = (stricmp(arg+9,MIN_CONNECTVER)>=0) && (stricmp(arg+9,MAX_CONNECTVER)<=0);
		}


		if (type == T_LOOKUP)
		{
			if (http.isHeader("remote-ip"))
			{
				if (!servMgr->forceIP[0])
				{
					Host nh;
					nh.fromStr(arg);
					if (nh.ip != servMgr->serverHost.ip)
					{
						LOG_DEBUG("Got new IP: %s",arg);
						servMgr->serverHost.ip = nh.ip;
					}
				}
			}else if (http.isHeader(PCX_HS_MSG))
			{
				if (strcmp(servMgr->rootMsg.cstr(),arg))
				{
					servMgr->rootMsg.set(arg);
					if (!servMgr->rootMsg.isEmpty())
						sys->setNotify(Sys::NT_PEERCAST,"Message from PeerCast",arg);
				}

			}else if (http.isHeader(PCX_HS_DL))
			{
				if (servMgr->downloadURL[0]==0)
				{
					strcpy(servMgr->downloadURL,arg);
					sys->setNotify(Sys::NT_UPGRADE,"PeerCast version update","There is a new version available, please click this icon to upgrade your client.");
				}
			}else if (http.isHeader(PCX_HS_FULLHIT))
				chanMgr->fullHitFrequency = atoi(arg);
			else if (http.isHeader(PCX_HS_BCTTL))
				chanMgr->broadcastTTL = atoi(arg);
		}

    }

	if ((stricmp(servMgr->network.cstr(),"peercast")!=0) && (!subnetValid))
		throw HTTPException(HTTP_SC_NOTFOUND,404);

	if (!versionValid)
		throw HTTPException(HTTP_SC_UNAUTHORIZED,401);

    sock->writeLine(GNU_OK);
    sock->writeLine("");
}
// -----------------------------------
void Servent::handshakeIn()
{
	sock->timeout = 10000;


	int osType=0;

	HTTP http(*sock);


	bool subnetValid = false;
	bool versionValid = false;
	bool diffRootVer = false;

    while (http.nextHeader())
    {
		char *arg = http.getArgStr();
		if (!arg)
			continue;

		if (http.isHeader(PCX_HS_SUBNET))
			subnetValid = stricmp(arg,servMgr->network)==0;
		else if (http.isHeader(HTTP_HS_AGENT))
		{
			agent.set(arg);

			if (strnicmp(arg,"PeerCast/",9)==0)
			{
				versionValid = (stricmp(arg+9,MIN_CONNECTVER)>=0) && (stricmp(arg+9,MAX_CONNECTVER)<=0);
				diffRootVer = stricmp(arg+9,MIN_ROOTVER)<0;
			}

		}else if (http.isHeader(PCX_HS_ID))
		{
			GnuID id;
			id.fromStr(arg);
			if (id.isSame(servMgr->sessionID))
				throw StreamException("Servent loopback");

		}else if (http.isHeader(PCX_HS_OS))
		{
			if (stricmp(arg,PCX_OS_LINUXDYN)==0)
				osType = 1;
			else if (stricmp(arg,PCX_OS_LINUXSTA)==0)
				osType = 2;
			else if (stricmp(arg,PCX_OS_WIN32)==0)
				osType = 3;
			else if (stricmp(arg,PCX_OS_MACOSX)==0)
				osType = 4;
		}

    }

	if ((stricmp(servMgr->network.cstr(),"peercast")!=0) && (!subnetValid))
		throw HTTPException(HTTP_SC_NOTFOUND,404);

	if (!versionValid)
	{
		throw HTTPException(HTTP_SC_UNAUTHORIZED,401);
	}


    sock->writeLine(GNU_OK);

    sock->writeLine("%s %s",HTTP_HS_AGENT,PCX_AGENT);
    sock->writeLine("%s %s",PCX_HS_SUBNET,servMgr->network.cstr());

	if (servMgr->isRoot)
	{
		sock->writeLine("%s %d",PCX_HS_BCTTL,9);

		if (diffRootVer)
		{
			sock->writeString(PCX_HS_DL);
			switch(osType)
			{
				case 1:
					sock->writeLine(PCX_DL_LINUXDYN);
					break;
				case 2:
					sock->writeLine(PCX_DL_LINUXSTA);
					break;
				case 3:
					sock->writeLine(PCX_DL_WIN32);
					break;
				case 4:
					sock->writeLine(PCX_DL_MACOSX);
					break;
				default:
					sock->writeLine(PCX_DL_URL);
					break;
			}
		}

		sock->writeLine("%s %s",PCX_HS_MSG,servMgr->rootMsg.cstr());
	}


	char hostIP[64];
	Host h = sock->host;
	h.IPtoStr(hostIP);
    sock->writeLine("Remote-IP: %s",hostIP);


    sock->writeLine("");


	while (http.nextHeader());

}

// -----------------------------------
void Servent::handshakeStream(Channel *ch)
{
	sock->timeout = 10000;

	HTTP http(*sock);

	while (http.nextHeader())
	{
		char *arg = http.getArgStr();
		if (!arg)
			continue;

		if (http.isHeader(HTTP_HS_AGENT))
			agent.set(arg);

		if (ch->info.type == ChanInfo::T_MP3)
			if (http.isHeader("icy-metadata"))
				addMetadata = atoi(arg) > 0;

		LOG_DEBUG("Stream: %s",http.cmdLine);
	}


	if (addMetadata)		// winamp mp3 metadata check
	{
	    sock->writeLine(ICY_OK);


		sock->writeLine("icy-name:%s",ch->getName());
		sock->writeLine("icy-br:%d",ch->getBitrate());
		sock->writeLine("icy-url:%s",ch->info.url.cstr());

		if (addMetadata)
			sock->writeLine("icy-metaint:%d",chanMgr->icyMetaInterval);

	    sock->writeLine("%s %s",HTTP_HS_CONTENT,MIME_MP3);

	}else
	{
	    sock->writeLine(HTTP_SC_OK);
		sock->writeLine("%s %s",HTTP_HS_SERVER,PCX_AGENT);
		//sock->writeLine(HTTP_LEN);

		if (ch->info.type == ChanInfo::T_MOV)
		{
			sock->writeLine("Accept-Ranges: bytes");
			sock->writeLine("Connection: close");
			sock->writeLine("Content-Length: 10000000");
		}

		switch (ch->info.type)
		{
			case ChanInfo::T_OGG:
			    sock->writeLine("%s %s",HTTP_HS_CONTENT,MIME_XOGG);
				break;
			case ChanInfo::T_MP3:
			    sock->writeLine("%s %s",HTTP_HS_CONTENT,MIME_MP3);
				break;
			case ChanInfo::T_MOV:
			    sock->writeLine("%s %s",HTTP_HS_CONTENT,MIME_MOV);
				break;
			case ChanInfo::T_MPG:
			    sock->writeLine("%s %s",HTTP_HS_CONTENT,MIME_MPG);
				break;
		}
	}


	
    sock->writeLine("");

}

// -----------------------------------
void Servent::handshakeGiv(Channel *ch)
{
	sock->timeout = 10000;

	char idstr[64];

	pushID.toStr(idstr);
    sock->writeLine("GIV %d:%s/%s",chanIndex,idstr,chanInfo.name.cstr());
	handshakeStream(ch);
}


// -----------------------------------
void Servent::process()
{
	try 
	{
	

		gnuStream.init(sock);
		setStatus(S_CONNECTED);


		sock->timeout = 10000;		// 10 seconds

		gnuStream.ping(2);

		if (type != T_LOOKUP)
			chanMgr->broadcastRelays(2,this);

		lastPacket = lastPing = sys->getTime();
		bool doneBigPing=false;

		const unsigned int	abortTimeoutSecs = 60;		// abort connection after 60 secs of no activitiy
		const unsigned int	packetTimeoutSecs = 30;		// ping connection after 30 secs of no activity

		unsigned int currBytes=0,lastTotal=0;
		unsigned int lastWait=0;

		while (thread.active && sock->active())
		{

			if (sock->readPending())
			{
				lastPacket = sys->getTime();
				if (gnuStream.readPacket(pack))
				{
					unsigned int ver = pack.id.getVersion();


					char ipstr[64];
					sock->host.toStr(ipstr);
					GnuStream::R_TYPE r;

					bool isOld=false;
					bool isDup = servMgr->seenPacket(pack);
					seenIDs.addGnuID(pack.id);


					servMgr->addVersion(ver);

					if ((ver < MIN_PACKETVER) || (ver >= MAX_PACKETVER))
						isOld = true;

					if (type == T_LOOKUP)
						isDup = false;

					LOG_NETWORK("packet in: %d - %d bytes - v%05x from %s %s %s",pack.func,pack.len,pack.id.getVersion(),ipstr,isDup?" - seen":"",isOld?"old":"");


					// if old packet, create unique ID based on packet data.. 
					//if (pack.id.getVersion() < 0x0106)
					//{
					//	pack.makeChecksumID();
					//	//LOG4("old: %d",pack.func);
					//}

					if ((!isDup) && (!isOld))
					{
						GnuID routeID;

						GnuPacket out;
						out.func = 255;
						r = gnuStream.processPacket(pack,out,this,routeID);
						if (out.func != 255)
							outputPacket(out);

						if (!pack.ttl)
						{
							stats.add(Stats::NUMDEAD);
						}else
						{
							int cnt;
							switch(r)
							{
								case GnuStream::R_BROADCAST:
									cnt = servMgr->broadcast(pack,this);
									if (!cnt)
										stats.add(Stats::NUMDROPPED);
									else
										stats.add(Stats::NUMBROADCASTED);

									break;
								case GnuStream::R_ROUTE:
									cnt = servMgr->route(pack,routeID,this);
									if (!cnt)
										stats.add(Stats::NUMDROPPED);
									else
										stats.add(Stats::NUMROUTED);

									break;
								case GnuStream::R_DISCARD:
									stats.add(Stats::NUMDISCARDED);
									break;
							}
						}

					}else
					{
						if (isDup)
							stats.add(Stats::NUMDUP);

						if (isOld)
							stats.add(Stats::NUMOLD);
					}




				}else{
					LOG_ERROR("Bad packet");
				}
			}


			if (readOut < writeOut)
			{
				GnuPacket *p = &outPack[readOut%MAX_OUTPACKETS];
				gnuStream.sendPacket(*p);
				seenIDs.addGnuID(p->id);
				readOut++;
			}

			int lpt =  sys->getTime()-lastPacket;

			if (!doneBigPing)
			{
				if ((sys->getTime()-lastPing) > 15)
				{
					gnuStream.ping(7);
					lastPing = sys->getTime();
					doneBigPing = true;
				}
			}else{
				if (lpt > packetTimeoutSecs)
				{
					
					if ((sys->getTime()-lastPing) > packetTimeoutSecs)
					{
						gnuStream.ping(1);
						lastPing = sys->getTime();
					}

				}
			}
			if (lpt > abortTimeoutSecs)
				throw TimeoutException();

			unsigned int tot = sock->totalBytesIn+sock->totalBytesOut;
			unsigned int bytes = (tot-lastTotal);
			lastTotal = tot;
			unsigned int delay = sys->idleSleepTime;
			if ((bytes) && (servMgr->serventBandwidth >= 8))
				delay = (bytes*1000)/(servMgr->serventBandwidth/8);

			if (delay < sys->idleSleepTime)
				delay = sys->idleSleepTime;

			sys->sleep(delay);

		}


	}catch(StreamException &e)
	{
		LOG_ERROR("Process Error: %s",e.msg);
	}

}
// -----------------------------------
void Servent::processRelay()
{
	try 
	{
	
		sock->timeout = 10000;		// 10 seconds

		gnuStream.init(sock);
		setStatus(S_CONNECTED);

		gnuStream.ping(2);

		while (thread.active && sock->active())
		{
			if (gnuStream.readPacket(pack))
			{
				char ipstr[64];
				sock->host.toStr(ipstr);
				unsigned int ver = pack.id.getVersion();

				servMgr->addVersion(ver);
				
				LOG_NETWORK("packet in: %d v%05x from %s",pack.func,pack.getVersion(),ipstr);

				//if (pack.id.getVersion() < MIN_PACKETVER)
				//	break;

				if (pack.func == 0)		// if ping then pong back some hosts and close
				{
					
					Host hl[64];
					int cnt = servMgr->getNewestHosts(hl,64,sock->host);	
					if (cnt)
					{
						int start = sys->rnd() % cnt;
						for(int i=0; i<8; i++)
						{
							GnuPacket pong;
							pack.hops = 2;
							pong.initPong(hl[start],false,pack);
							gnuStream.sendPacket(pong);

							start = (start+1) % cnt;
						}
						char str[64];
						sock->host.toStr(str);
						LOG_NETWORK("Sent 8 pongs to %s",str);
					}else
					{
						LOG_NETWORK("No Pongs to send");
						return;
					}
				}else if (pack.func == 1)		// pong?
				{
					MemoryStream pong(pack.data,pack.len);

					int ip,port;
					port = pong.readShort();
					ip = pong.readLong();
					ip = SWAP4(ip);

					LOG_NETWORK("pong: %d.%d.%d.%d:%d",ip>>24&0xff,ip>>16&0xff,ip>>8&0xff,ip&0xff,port);

					if ((ip) && (port))
					{
						Host h(ip,port);
						servMgr->addHost(h,true);
					}
					return;
				}

				if (gnuStream.packetsIn > 5)	// die if we get too many packets
					return;
			}
		}


	}catch(StreamException &e)
	{
		LOG_ERROR("Relay: %s",e.msg);
	}



}

// -----------------------------------
int Servent::givProc(ThreadInfo *thread)
{
	thread->lock();
	Servent *sv = (Servent*)thread->data;
	try 
	{
		sv->processStream(true);
	}catch(StreamException &e)
	{
		LOG_ERROR("GIV: %s",e.msg);
	}

	sv->endThread();
	return 0;
}
// -----------------------------------
int Servent::outgoingProc(ThreadInfo *thread)
{
	thread->lock();

	Servent *sv = (Servent*)thread->data;
		
	try 
	{
		sv->setStatus(S_CONNECTING);
		sv->sock->timeout = 10000;		// 10 seconds handshake
		sv->sock->connect();
		sv->setStatus(S_HANDSHAKE);
		sv->handshakeOut();
		sv->process();

	}catch(TimeoutException &e)
	{
		LOG_ERROR("Outgoing Timeout: %s",e.msg);
		servMgr->deadHost(sv->sock->host);
		sv->setStatus(S_TIMEOUT);
	}catch(StreamException &e)
	{
		servMgr->deadHost(sv->sock->host);
		LOG_ERROR("Outgoing: %s",e.msg);
		sv->setStatus(S_REFUSED);

		if (sv->type == T_LOOKUP)
			sys->sleep(30000);
	}


	sv->endThread();
	return 0;
}
// -----------------------------------
int Servent::incomingProc(ThreadInfo *thread)
{
	thread->lock();

	Servent *sv = (Servent*)thread->data;
	
	try 
	{
		sv->handshakeIncoming();
	}catch(HTTPException &e)
	{
		try
		{
			sv->sock->writeLine(e.msg);
			if (e.code == 401)
				sv->sock->writeLine("WWW-Authenticate: Basic realm=\"PeerCast\"");
			sv->sock->writeLine("");
		}catch(StreamException &){}
		LOG_ERROR("Incoming HTTP error: %s",e.msg);
	}catch(StreamException &e)
	{
		LOG_ERROR("Incoming Connect error: %s",e.msg);
	}

	sv->endThread();
	return 0;
}
// -----------------------------------
void Servent::processServent()
{
	setStatus(S_HANDSHAKE);

	handshakeIn();

	if (!sock)
		throw StreamException("Servent has no socket");


	if (servMgr->isRoot && !servMgr->needConnections())
		processRelay();
	else
		process();

}

// -----------------------------------
void Servent::processStream(bool push)
{
		
	Channel *ch = NULL;
	try 
	{
		setStatus(S_HANDSHAKE);


		ch = chanMgr->findChannel(chanInfo);
		if (!ch)
			throw StreamException("Not found");

		chanIndex = ch->index;

		if (push)
			handshakeGiv(ch);
		else
			handshakeStream(ch);

		servMgr->totalStreams++;

		if (isRaw)
		{
			if (addMetadata)
				sendRawMetaChannel(ch,chanMgr->icyMetaInterval);
			else
				sendRawChannel(ch);
		}else
			sendChannel(ch);

		setStatus(S_CLOSING);

	}catch(StreamException &e)
	{
		setStatus(S_ERROR);
		LOG_ERROR("Stream Error: %s",e.msg);
	}
}

// -----------------------------------------
#if 0
// debug
		FileStream file;
		file.openReadOnly("c://test.mp3");

		LOG_DEBUG("raw file read");
		char buf[4000];
		int cnt=0;
		while (!file.eof())
		{
			LOG_DEBUG("send %d",cnt++);
			file.read(buf,sizeof(buf));
			sock->write(buf,sizeof(buf));

		}
		file.close();
		LOG_DEBUG("raw file sent");

	return;
// debug
#endif
// -----------------------------------
bool Servent::waitForChannelHeader(Channel *ch)
{
	for(int i=0; i<60; i++)
	{
		if (ch->isPlaying() && ch->syncPos)
			return true;

		if (!thread.active || !sock->active())
			break;
		sys->sleep(1000);
	}
	return false;
}

// -----------------------------------
void Servent::sendRawChannel(Channel *ch)
{

	ch->numListeners++;
	outputBitrate = ch->getBitrate();

	try
	{

		sock->timeout = 10000;
		setStatus(S_CONNECTED);

		LOG_DEBUG("Starting Raw stream: %s",chanInfo.name.cstr());

		if (!waitForChannelHeader(ch))
			throw StreamException("Channel not ready");


		if (ch->headMeta.len)
		{
			LOG_DEBUG("Send %d bytes header ",ch->headMeta.len);
			sock->write(&ch->headMeta.data,ch->headMeta.len);
		}

		currPos=0;
		unsigned int mySync=0,chanSync=0;

		unsigned int syncTimeout=0;

		ChanPacket pack;

		while ((thread.active) && sock->active())
		{
			if (!ch->isActive())
				throw StreamException("Channel closed");

			currPos = ch->chanData.readPacket(currPos,pack);

			//LOG_CHANNEL("Send POS: %d",currPos);

			if (pack.type == 'SYNC')
			{
				chanSync = *(unsigned int *)pack.data;
				//LOG_CHANNEL("Send SYNC: %d",chanSync);
				int rs = mySync-chanSync;
				if (rs > ChanPacketBuffer::MAX_PACKETS)
					mySync = chanSync-1;
			}


			if (pack.type == 'DATA')
			{
				// skip packets until we`re back in sync
				if (chanSync <= mySync)
				{
					syncTimeout++;
					if (syncTimeout > 30)
						throw StreamException("Unable to sync");
					LOG_CHANNEL("SYNC wait: %d - %d",mySync-chanSync,syncTimeout);
				}else
				{
					//LOG_CHANNEL("Send DATA: %d",chanSync);
					syncTimeout=0;
					mySync = chanSync;
					sock->write(pack.data,pack.len);
				}
			}

		}
	}catch(StreamException &e)
	{
		LOG_ERROR("Stream channel: %s",e.msg);
	}
	if (ch->numListeners)
		ch->numListeners--;
}

// -----------------------------------
void Servent::sendRawMetaChannel(Channel *ch, int interval)
{
	ch->numListeners++;
	outputBitrate = ch->getBitrate();

	try
	{

		sock->timeout = 10000;
		setStatus(S_CONNECTED);

		LOG_DEBUG("Starting Raw Meta stream: %s",chanInfo.name.cstr());

		if (!waitForChannelHeader(ch))
			throw StreamException("Channel not ready");


		if (ch->headMeta.len)
		{
			LOG_DEBUG("Send %d bytes header ",ch->headMeta.len);
			sock->write(&ch->headMeta.data,ch->headMeta.len);
		}

		String lastTitle,lastURL;

		currPos=0;
		unsigned int mySync=0,chanSync=0;

		int		lastMsgTime=sys->getTime();
		bool	showMsg=true;

		ChanPacket pack;

		char buf[16384];
		int bufPos=0;

		if ((interval > sizeof(buf)) || (interval < 1))
			throw StreamException("Bad ICY Meta Interval value");

		unsigned int syncTimeout=0;

		while ((thread.active) && sock->active())
		{
			if (!ch->isActive())
				throw StreamException("Channel closed");

			currPos = ch->chanData.readPacket(currPos,pack);
			MemoryStream mem(pack.data,pack.len);

			if (pack.type == 'SYNC')
			{
				chanSync = mem.readLong();
				int rs = mySync-chanSync;
				if (rs > ChanPacketBuffer::MAX_PACKETS)
					mySync = chanSync-1;
			}

			if (pack.type == 'DATA')
			{
				// skip packets until we`re back in sync
				if (chanSync <= mySync)
				{
					syncTimeout++;
					if (syncTimeout > 30)
						throw StreamException("Unable to sync");
					LOG_DEBUG("SYNC wait: %d - %d",mySync-chanSync,syncTimeout);
				}else
				{
					syncTimeout = 0;
					mySync = chanSync;

					int len = pack.len;
					char *p = pack.data;
					while (len)
					{
						int rl = len;
						if ((bufPos+rl) > interval)
							rl = interval-bufPos;
						memcpy(&buf[bufPos],p,rl);
						bufPos+=rl;
						p+=rl;
						len-=rl;

						if (bufPos >= interval)
						{
							bufPos = 0;	
							sock->write(buf,interval);

							if (chanMgr->broadcastMsgInterval)
								if ((sys->getTime()-lastMsgTime) >= chanMgr->broadcastMsgInterval)
								{
									showMsg ^= true;
									lastMsgTime = sys->getTime();
								}

							String *metaTitle = &ch->info.track.title;
							if (!ch->info.comment.isEmpty() && (showMsg))
								metaTitle = &ch->info.comment;


							if (!metaTitle->isSame(lastTitle) || !ch->info.url.isSame(lastURL))
							{

								char tmp[1024];
								String title,url;

								title = *metaTitle;
								url = ch->info.url;

								title.convertTo(String::T_META);
								url.convertTo(String::T_META);

								sprintf(tmp,"StreamTitle='%s';StreamUrl='%s';\0",title.cstr(),url.cstr());
								int len = ((strlen(tmp) + 15+1) / 16);
								sock->writeChar(len);
								sock->write(tmp,len*16);

								lastTitle = *metaTitle;
								lastURL = ch->info.url;

								LOG_DEBUG("StreamTitle: %s, StreamURL: %s",lastTitle.cstr(),lastURL.cstr());

							}else
								sock->writeChar(0);					

						}
					}
				}
			}
		}
	}catch(StreamException &e)
	{
		LOG_ERROR("Stream channel: %s",e.msg);
	}
	if (ch->numListeners)
		ch->numListeners--;
}
// -----------------------------------
void Servent::sendChannel(Channel *ch)
{
	ch->numRelays++;
	outputBitrate = ch->getBitrate();

	try
	{


		sock->timeout = 10000;
		setStatus(S_CONNECTED);

		LOG_DEBUG("Starting PeerCast stream: %s",chanInfo.name.cstr());


		currPos=0;

		sock->writeTag("PCST");

		ChanPacket pack;

		pack.init('HEAD',ch->headMeta.data,ch->headMeta.len);
		pack.write(*sock);

		pack.init('META',ch->insertMeta.data,ch->insertMeta.len);
		pack.write(*sock);
		

		while ((thread.active) && sock->active())
		{
			if (!ch->isActive())
				throw StreamException("Channel closed");

			unsigned int np = ch->chanData.readPacket(currPos,pack);
			if ((np-currPos) > 1)
				LOG_DEBUG("sendChannel skip: %d",np-currPos);
			currPos = np;

			pack.write(*sock);
		}

	}catch(StreamException &e)
	{
		LOG_ERROR("Stream channel: %s",e.msg);
	}

	if (ch->numRelays)
		ch->numRelays--;
}

// -----------------------------------
int Servent::serverProc(ThreadInfo *thread)
{
	thread->lock();

	Servent *sv = (Servent*)thread->data;

	try 
	{
		if (!sv->sock)
			throw StreamException("Server has no socket");

		sv->setStatus(S_LISTENING);
		//LOG4("Listening on port %d",sv->sock->host.port);


		char servIP[64];
		sv->sock->host.toStr(servIP);

		if (servMgr->isRoot)
			LOG_DEBUG("Root Server started: %s",servIP);
		else
			LOG_DEBUG("Server started: %s",servIP);
		

		while ((thread->active) && (sv->sock->active()))
		{
			ClientSocket *cs = sv->sock->accept();
			if (cs)
			{				
				// if global IP then we`re not firewalled
				if (cs->host.globalIP())
					servMgr->setFirewall(ServMgr::FW_OFF);

				Servent *ns = servMgr->allocServent();
				if (ns)
					ns->initIncoming(cs,sv->allow);
				else
					LOG_ERROR("Out of servents");
			}
		}
	}catch(StreamException &e)
	{
		LOG_ERROR("Server Error: %s:%d",e.msg,e.err);
	}

	
	LOG_DEBUG("Server stopped");

	sv->endThread();
	return 0;
}
 
